from typing import Optional, List, Tuple, Callable

# Get a filename format
def get_format (filename : str) -> str:
    return filename.split('.')[-1]

# Find a function which is suitable for any of the available request "format sets"**
# All functions are checked for each request format set before jumping to another and they are evaluated in order
# A function and new generated format set with formats in common are returned
# None is returned when there is no suitable function
# WARNING: Available functions must have the 'format_sets' property
# ** Format sets are dictionaries which specify input and output formats
# Consider function format sets as 'required input' and 'available output'
# Consider request format sets as 'available input' and 'required output'
# Both inputs and ouputs are dictionaries where keys are function keywords and values are sets of supported formats
# Alternatively, a keyword may have None as value to represent unnecessary requirements or missing availabilities
# An example is shown below:
# {
#     'inputs': {
#         'input_structure_filename': {'tpr'},
#         'input_trajectory_filename': {'xtc', 'trr'},
#     },
#     'outputs': {
#         'output_trajectory_filename': {'pdb', 'gro'}
#     },
# }
# Functions may have multiple format sets since different input formats may lead to different output formats
def get_format_set_suitable_function (
    available_functions : List[Callable],
    available_request_format_sets : List[dict],
) -> Optional[ Tuple[ Callable, dict ] ]:
    # Try with each request format set
    for request_format_set in available_request_format_sets:
        # Search functions to match formats for every keyword
        for function in available_functions:
            # Test every function format set independently
            for function_format_set in function.format_sets:
                # Get common values for every keyword format set
                common_format_set = {'inputs':{},'outputs':{}}
                missing_common_format = False
                # Check format keys are compatible
                if not check_format_sets_compability(request_format_set, function_format_set):
                    raise ValueError('Format keys are not compatible with function ' + str(function.__name__))
                # Check the function inputs to be fulfilled by the request inputs
                required_input_keywords = function_format_set['inputs'].keys()
                for keyword in required_input_keywords:
                    required_formats = function_format_set['inputs'][keyword]
                    # If this format set does not need an input we set None for this keyword
                    if required_formats == None:
                        common_format_set[keyword] = None
                        continue
                    available_formats = request_format_set['inputs'][keyword]
                    # If this format set is missing an input we are done
                    if available_formats == None:
                        missing_common_format = True
                        break
                    common_formats = available_formats.intersection(required_formats)
                    # If the common formats set is empty we are done
                    if not common_formats:
                        missing_common_format = True
                        break
                    common_format_set['inputs'][keyword] = common_formats
                # If any of the common format sets was empty it means formats do not match
                if missing_common_format:
                    continue
                # Check the request outputs to be fulfilled by the function outputs
                required_outpt_keywords = request_format_set['outputs'].keys()
                for keyword in required_outpt_keywords:
                    required_formats = request_format_set['outputs'][keyword]
                    # If this format set does not need an input we set None for this keyword
                    if required_formats == None:
                        common_format_set[keyword] = None
                        continue
                    available_formats = function_format_set['outputs'][keyword]
                    # If this format set is missing an input we are done
                    if available_formats == None:
                        missing_common_format = True
                        break
                    common_formats = available_formats.intersection(required_formats)
                    # If the common formats set is empty we are done
                    if not common_formats:
                        missing_common_format = True
                        break
                    common_format_set['outputs'][keyword] = common_formats
                # If any of the common format sets was empty it means formats do not match
                if missing_common_format:
                    continue
                # Otherwise we have the function
                return function, common_format_set
                

# Check two format sets to be compatible
# Both function and request format sets must match in their requirements
# i.e. all function format set input keywords must be included in request format set input keywords
# i.e. all request format set output keywords must be included in function format set output keywords
def check_format_sets_compability (request_format_set : dict, function_format_set : dict) -> bool:
    # Check the function inputs keyowrds to exist in the request input keywords
    required_input_keywords = function_format_set['inputs'].keys()
    available_input_keywords = request_format_set['inputs'].keys()
    for keyword in required_input_keywords:
        if keyword not in available_input_keywords:
            print('ERROR: Missing ' + keyword + ' keyword')
            return False
    # Check the request output keyowrds to exist in the function output keywords
    required_outpt_keywords = request_format_set['outputs'].keys()
    available_outpt_keywords = function_format_set['outputs'].keys()
    for keyword in required_outpt_keywords:
        if keyword not in available_outpt_keywords:
            print('ERROR: Missing ' + keyword + ' keyword')
            return False
    return True




# Structure file formats

def is_pdb (filename : str) -> bool:
    return filename[-4:] == '.pdb'

def is_psf (filename : str) -> bool:
    return filename[-4:] == '.psf'

def is_tpr (filename : str) -> bool:
    return filename[-4:] == '.tpr'

def is_gro (filename : str) -> bool:
    return filename[-4:] == '.gro'

def is_prmtop (filename : str) -> bool:
    return filename[-7:] == '.prmtop'

def is_top (filename : str) -> bool:
    return filename[-4:] == '.top'

# Trajectory file formats

def is_xtc (filename : str) -> bool:
    return filename[-4:] == '.xtc'

def is_dcd (filename : str) -> bool:
    return filename[-4:] == '.dcd'

def is_netcdf (filename : str) -> bool:
    return filename[-3:] == '.nc'

def are_xtc (filenames : list) -> bool:
    return all([ is_xtc(filename) for filename in filenames ])

def are_dcd (filenames : list) -> bool:
    return all([ is_dcd(filename) for filename in filenames ])

def are_netcdf (filenames : list) -> bool:
    return all([ is_netcdf(filename) for filename in filenames ])

# Extra formats logic

# Check if a file may be read by pytraj according to its format
def is_pytraj_supported (filename : str) -> bool:
    return is_prmtop(filename) or is_top(filename) or is_psf(filename)

# From GitHub:
# ParmFormatDict = {
#     "AMBERPARM": AMBERPARM,
#     "PDBFILE": PDBFILEPARM,
#     "MOL2FILE": MOL2FILEPARM,
#     "CHARMMPSF": CHARMMPSF,
#     "CIFFILE": CIFFILE,
#     "GMXTOP": GMXTOP,
#     "SDFFILE": SDFFILE,
#     "TINKER": TINKERPARM,
#     "UNKNOWN_PARM": UNKNOWN_PARM,
# }

# Get the pytraj format key for the write_parm function for a specific file according to its format
def get_pytraj_parm_format (filename : str) -> str:
    if is_prmtop(filename):
        return 'AMBERPARM'
    if is_psf(filename):
        return 'CHARMMPSF'
    if is_top(filename):
        return 'GMXTOP'
    if is_pdb(filename):
        return 'PDBFILE'
    raise ValueError('The file ' + filename + ' format is not supported')


# Subset of tools used alone and directly from the console

import sys
import os
from subprocess import run, PIPE

from typing import Optional

# This tool allows you to set the chain of all atoms in a selection
# This is powered by VMD and thus the selection lenguage must be the VMD's
# Arguments are as follows:
# 1 - Input pdb filename
# 2 - Atom selection (All atoms by defualt)
# 3 - Chain letter (May be the flag 'fragment', which is the default indeed)
# 4 - Output pdb filename (Input filename by default)
def chainer (
    input_pdb_filename : str,
    atom_selection : Optional[str] = None,
    chain_letter : Optional[str] = None,
    output_pdb_filename : Optional[str] = None
):

    # If no atom selection is provided then all atoms are selected
    if not atom_selection:
        atom_selection = 'all'

    # If no chain letter is provided then the flag 'fragment' is used
    if not chain_letter:
        chain_letter = 'fragment'

    # If no output filename is provided then use input filename as output filename
    if not output_pdb_filename:
        output_pdb_filename = input_pdb_filename

    # Check the file exists
    if not os.path.exists(input_pdb_filename):
        raise SystemExit('ERROR: The file does not exist')

    # Set he path to a script with all commands needed for VMD to parse the topology file
    # The scripts is in Tcl lenguage
    commands_filename = '.commands.vmd'
       
    with open(commands_filename, "w") as file:
        # Select the specified atoms and set the specified chain
        file.write('set atoms [atomselect top "' + atom_selection + '"]\n')
        # In case chain letter is not a letter but the 'fragment' flag, asign chains by fragment
        # Fragments are atom groups which are not connected by any bond
        if chain_letter == 'fragment':
            # Get all different chain names
            file.write('set chains_sample [lsort -unique [${atoms} get chain]]\n')
            # Set letters in alphabetic order
            file.write('set letters "A B C D E F G H I J K L M N O P Q R S T U V W X Y Z"\n')
            # Get the number of fragments
            file.write('set fragment_number [llength [lsort -unique -integer [${atoms} get fragment]]]\n')
            # For each fragment, set the chain of all atoms which belong to this fragment alphabetically
            # e.g. fragment 0 -> chain A, fragment 1 -> chain B, ...
            file.write('for {set i 0} {$i <= $fragment_number} {incr i} {\n')
            file.write('	set fragment_atoms [atomselect top "fragment $i"]\n')
            file.write('	$fragment_atoms set chain [lindex $letters $i]\n')
            file.write('}\n')
            # Otherwise, set the specified chain
        else:
            file.write('$atoms set chain ' + chain_letter + '\n')
        # Write the current topology in 'pdb' format
        file.write('set all [atomselect top "all"]\n')
        file.write('$all frame first\n')
        file.write('$all writepdb ' + output_pdb_filename + '\n')
        file.write('exit\n')

    # Run VMD
    logs = run([
        "vmd",
        input_pdb_filename,
        "-e",
        commands_filename,
        "-dispdev",
        "none"
    ], stdout=PIPE).stdout.decode()

    # Remove the vmd commands file
    os.remove(commands_filename)


# Functions powered by VMD
# Humphrey, W., Dalke, A. and Schulten, K., "VMD - Visual Molecular Dynamics", J. Molec. Graphics, 1996, vol. 14, pp. 33-38. 
# http://www.ks.uiuc.edu/Research/vmd/

import os

from subprocess import run, PIPE, Popen

from .single_frame_getter import get_first_frame

# Set the script filename with all commands to be passed to vmd
commands_filename = '.commands.vmd'

# List all the vmd supported trajectory formats
vmd_supported_trajectory_formats = {'mdcrd', 'crd', 'dcd', 'xtc', 'trr'}

# Given a vmd supported topology with no coordinates and a single frame file, generate a pdb file
def vmd_to_pdb (
    input_structure_filename : str,
    input_trajectory_filename : str,
    output_structure_filename : str):

    single_frame_filename = get_first_frame(input_trajectory_filename, vmd_supported_trajectory_formats)

    # Prepare a script for VMD to run. This is Tcl language
    with open(commands_filename, "w") as file:
        # Select all atoms
        file.write('set all [atomselect top "all"]\n')
        # Write the current topology in 'pdb' format
        file.write('$all frame first\n')
        file.write('$all writepdb ' + output_structure_filename + '\n')
        file.write('exit\n')

    # Run VMD
    logs = run([
        "vmd",
        input_structure_filename,
        single_frame_filename,
        "-e",
        commands_filename,
        "-dispdev",
        "none"
    ], stdout=PIPE).stdout.decode()

    os.remove(commands_filename)
# Set function supported formats
vmd_to_pdb.format_sets = [
    {
        'inputs': {
            'input_structure_filename': {'psf', 'parm'},
            'input_trajectory_filename': vmd_supported_trajectory_formats
        },
        'outputs': {
            'output_structure_filename': {'pdb'}
        }
    }
]


import os
from subprocess import run, PIPE

# Multiple files may be selected with bash syntax (e.g. *.dcd)
# Tested supported input formats are .dcd
# Tested supported output formats are .xtc
def merge_and_convert_traj (
    input_filenames : list,
    output_filename : str
    ):

    # Run MDtraj
    logs = run([
        "mdconvert",
        "-o",
        output_filename,
        *input_filenames,
    ], stderr=PIPE).stderr.decode()
    # If output has not been generated then warn the user
    if not os.path.exists(output_filename):
        print(logs)
        raise SystemExit('Something went wrong with MDTraj')

def convert_traj (input_trajectory_filename : str, output_trajectory_filename : str):
    merge_and_convert_traj([input_trajectory_filename], output_trajectory_filename)
convert_traj.format_sets = [
    {
        'inputs': {
            'input_trajectory_filename': {'dcd', 'xtc', 'trr', 'nc', 'h5', 'binpos'}
        },
        'outputs': {
            'output_trajectory_filename': {'dcd', 'xtc', 'trr', 'nc', 'h5', 'binpos'}
        }
    },
]